#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
"""
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm
import matplotlib.ticker as mtick
from matplotlib.ticker import ScalarFormatter
import seaborn as sns
from numpy import nan_to_num

default_cmap = 'viridis'

__all__ = ["heatmap", "histplot"]


def heatmap(X, Y, Z, ax=None, logy=True, cbar=True, hide_low=True, 
            cmap=default_cmap, fig_kws={}, cbar_kws={}, plot_kws={}, 
            **kwargs):
    """Plot the heatmap of the particle size distribution. 
    All NaN'd values will be converted to zeroes.
    
    :param X: The boundaries of cells in x-axis dimension.
    :type X: array-like
    :param Y: The boundaries of cells in y-axis dimension.
    :type Y: array-like
    :param Z: The color value in the cells.
    :type Z: 2D array-like
    :param ax: A `matplotlib.pyplot` Axes in which to build the 
        plot. If `None`, a new Figure is instantiated. Defaults to 
        `None`.
    :type ax: `matplotlib.pyplot` Axes
    :param logy: Set whether the y axis is on a log scale, 
        defaults to True.
    :type logy: bool
    :param cbar: Set whether to include a color bar in the plot key, 
        defaults to True.
    :type cbar: bool
    :param hide_low: Set whether `Z` values below `cbar_min` should 
        be set to `cbar_min`, defaults to True.
    :type hide_low: bool
    :param cmap: A color map for `matplotlib.pyplot` `pcolormesh`, 
        defaults to `default_cmap`. See `here 
        <https://matplotlib.org/stable/api/_as_gen
        /matplotlib.colors.Colormap.html
        #matplotlib.colors.Colormap>`__.
    :type cmap: string or Colormap
    :param fig_kws: Options for the `matplotlib.pyplot` figure.
        See `here <https://matplotlib.org/stable/api/_as_gen
        /matplotlib.pyplot.figure.html>`__,
        efaults to `dict(figsize=(16,8)`.
    :type fig_kws: dict
    :param cbar_kws: Options for the `matplotlib.pyplot` `colorbar`.
        See `here <https://matplotlib.org/stable
        /api/figure_api.html#matplotlib.figure.Figure.colorbar>`__.
    :type cbar_kws: dict
    :param plot_kws: Options for the `matplotlib.pyplot` Axes.
        See `here <https://matplotlib.org/stable/api/_as_gen
        /matplotlib.pyplot.gca.html>'__,
        Defaults to `dict(alpha=1, edgecolor=None, linewidth=0)`.
    :type plot_kws: dict
    :param cbar_min: The minimum value for the color bar, 
        defaults to the smallest `Z` value
        (unless it is not greater than zero, 
        and then it will default to one).
    :type cbar_min: float
    :param cbar_max: The maximum value for the color bar, 
        defaults to the largest `Z` value.
    :type cbar_max: float
    :return: A plot of the histogram.
    :rtype: `matplotlib.pyplot` Axes
    """
    # Copy to avoid modifying original data
    Z_plot = Z.copy()

    # get rid of NaNs
    Z_plot = nan_to_num(Z_plot)

    # Set the colorbar min and max based on the min and 
    # max of the values
    cbar_min = kwargs.pop(
        'cbar_min', 
        Z_plot.min() if Z_plot.min() > 0.0 else 1.
    )
    cbar_max = kwargs.pop('cbar_max', Z_plot.max())

    if hide_low:
        # Increase values below cbar_min to cbar_min
        below_min = Z_plot < cbar_min
        Z_plot[below_min] = cbar_min

    # Set the plot_kws
    plot_kws = dict(
        dict(norm=LogNorm(vmin=cbar_min, vmax=cbar_max), cmap=cmap), 
        **plot_kws
    )

    # Set the figure keywords
    fig_kws = dict(dict(figsize=(10,5)), **fig_kws)

    if ax is None:
        plt.figure(**fig_kws)
        ax = plt.gca()

    # Plot the data as a pcolormesh
    im = ax.pcolormesh(X, Y, Z_plot, shading='auto', **plot_kws)

    # Set the ylim to match the data
    ax.set_ylim([Y.min(), Y.max()])

    # Set the axis to be log in the y-axis
    if logy:
        ax.semilogy()

        ax.yaxis.set_major_formatter(ScalarFormatter())

    ax.set_ylabel(r"$D_p\;[\mu m]$")

    if cbar:
        # Set the figure keywords
        cbar_kws = dict(
            dict(label=r'$dN/dlogD_p\;[cm^{-3}]$'), 
            **cbar_kws
        )

        clb = plt.colorbar(im, **cbar_kws)

    return ax


def histplot(histogram, bins, ax=None, plot_kws=None, fig_kws=None, 
             **kwargs):
    """
    Plot the histogram in the form of a bar chart.
    
    :param histogram: A histogram indexed by bins.
        If a DataFrame is given the data will be averaged over time.
    :type histogram: Pandas DataFrame, Pandas Series
    :param bins: For each bin a row with the lower boundary, the
        midpoint, and the upper bounary for that bin.
    :type bins: numpy array
    :param ax: `matplotlib.pyplot` Axes in which to build the plot.
        If `None`, a new Figure is instantiated.
    :type ax: `matplotlib.pyplot` Axes
    :param plot_kws: Options for the `matplotlib.pyplot` Axes.
        See `here <https://matplotlib.org/stable/api/_as_gen
        /matplotlib.pyplot.gca.html>'__.
        Defaults to `dict(alpha=1, edgecolor=None, linewidth=0)`.
    :type plot_kws: dict
    :param fig_kws: Options for the `matplotlib.pyplot` figure.
        See `here <https://matplotlib.org/stable/api/_as_gen
        /matplotlib.pyplot.figure.html>`__,
        defaults to `dict(figsize=(16,8)`.
    :type fig_kws: dict
    :return: A plot of the histogram.
    :rtype: `matplotlib.pyplot` Axes
    """
    if isinstance(histogram, pd.DataFrame):
        histogram = histogram.mean().values

    if fig_kws is None:
        fig_kws = dict(figsize=(16,8))

    if plot_kws is None:
        plot_kws = dict(alpha=1, edgecolor=None, linewidth=0)

    if ax is None:
        plt.figure(**fig_kws)
        ax = plt.gca()

    ax.bar(x=bins[:, 0], height=histogram, width=bins[:, -1] - bins[:, 0],
            align='edge', **plot_kws)

    ax.semilogx()

    ax.set_xlabel(r"$D_p\;[\mu m]$")

    ax.xaxis.set_major_formatter(mtick.FormatStrFormatter("%.3g"))

    # create a large number formatter for the y axis
    fmt = mtick.ScalarFormatter()
    fmt.set_powerlimits((-3, 2))

    ax.yaxis.set_major_formatter(fmt)

    return ax



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
"""
import pandas as pd
import numpy as np
import math
import copy
import joblib

from .utils import _get_bin_count, _get_linecount, make_bins
from .plots import heatmap
from .models import SMPS


__all__ = ["smps_from_txt", "load_sample"]


def smps_from_txt(fpath, column=True, delimiter=',', 
                  as_dict=True, **kwargs):
    """Read an SMPS txt file as exported by the TSI AIM software.
    
    :param fpath: The file path to txt file to be read.
    :type fpath: string
    :param column: If your data is in 'column' format, set True. Otherwise, set False, 
        defaults to True.
    :type column: bool
    :param delimiter: The delimiter in the txt file, 
        must be a comma of a tab, defaults to ",".
    :type delimiter: string
    :param as_dict: Sets the data type returned by the function, 
        defaults to True.
    :type as_dict: bool
    :return: The data loaded from the txt file.
    :rtype: SMPS instance or dict, depending on 
        the parameter as_dict
    """
    assert(delimiter in [',', '\t']), "The delimiter must \
    be either a comma or tab"

    encoding = kwargs.pop("encoding", "ISO-8859-1")

    # determine the number of rows of meta information; can be defined
    meta_num_lines = kwargs.pop("meta_num_lines", None)
    if not meta_num_lines:
        meta_num_lines = _get_linecount(
            fpath, 
            keyword='Sample #', 
            encoding=encoding, 
            delimiter=delimiter
        )

    # read and store the meta information
    meta = pd.read_table(
        fpath, 
        nrows=meta_num_lines, 
        delimiter=delimiter, 
        header=None, 
        encoding=encoding,
        on_bad_lines='warn',
        index_col=0
    ).T.iloc[0,:].to_dict()

    # grab the number of channels per decade - sets the 
    # multiplier for determining bin spacings later on
    mult = float(meta['Channels/Decade'])

    # determine the weight and units
    units = meta['Units'].lower()
    weight = meta['Weight'].lower()

    # read the data
    if column:
        ts = pd.read_table(
            fpath,
            skiprows=meta_num_lines,
            nrows=3,
            delimiter=delimiter,
            header=None,
            encoding=encoding).iloc[1:, 1:].T

        ts.columns = ['Date', 'Time']

        # set the timestamp
        ts['timestamp'] = ts.apply(
            lambda x: pd.to_datetime(
                "{} {}".format(x['Date'], 
                               x['Time']
                              )
            ), axis=1
        )

        # Retrieve the number of bins in the file
        nbins = _get_bin_count(
            fpath=fpath, 
            delimiter=delimiter, 
            encoding=encoding
        )

        # Read the table of raw data
        data = pd.read_table(
                    fpath,
                    skiprows=meta_num_lines+4,
                    nrows=nbins,
                    delimiter=delimiter,
                    header=None,
                    encoding=encoding).astype(float)

        midpoints = data.iloc[:, 0].values
        data = data.iloc[:, 1:].T

        # Alter the column names
        data.columns = ['bin{}'.format(i) for i in range(nbins)]
        data.index = ts['timestamp']

        # Read the table of stats
        stats = pd.read_table(
                    fpath,
                    skiprows=(meta_num_lines+4+nbins),
                    delimiter=delimiter,
                    header=None,
                    on_bad_lines='warn',
                    encoding=encoding).T

        # Rename the columns to the first row
        stats = stats.rename(columns=stats.iloc[0])

        # Drop the first row
        stats = stats.iloc[1:,:]

        # Set the index to timestamp
        stats.index = ts['timestamp']

        # Force the stats columns to be floats where needed
        for column in stats.columns:
            try:
                stats[column] = stats[column].astype(float)
            except: pass

        bin_labels = list(data.columns)

        data = pd.merge(
            data, 
            stats, 
            left_index=True, 
            right_index=True, 
            how='outer'
        )
    else:
        data = pd.read_table(
                    fpath,
                    skiprows=meta_num_lines,
                    delimiter=delimiter,
                    encoding=encoding).dropna(how='all', axis=1)

        data.index = data.apply(
            lambda x: pd.to_datetime("{} {}".format(
                x['Date'], 
                x['Start Time']
            )), axis=1
        )

        # delete some un-needed columns
        del data['Date'], data['Start Time']

        # grab the midpoint diameters
        midpoints = []
        for col in data.columns:
            try:
                float(col)
                midpoints.append(col)
            except ValueError:
                pass

        bin_labels = ["bin{}".format(i) for i in range(len(midpoints))]

        # rename the bins
        data.rename(
            columns=dict(zip(midpoints, bin_labels)), 
            inplace=True
        )

        midpoints = np.array([float(i) for i in midpoints])

    # generate the bins
    low_col_name = kwargs.pop(
        "lower_column_label", 
        [c for c in data.columns if "lower" in c.lower()][0]
    )
    upper_col_name = kwargs.pop(
        "upper_column_label", 
        [c for c in data.columns if "upper" in c.lower()][0]
    )
    bound_left = float(data[low_col_name][0])
    bound_right = float(data[upper_col_name][0])

    meta['Lower Size (nm)'], meta['Upper Size (nm)'] \
    = bound_left, bound_right

    # make the bin array
    bins = make_bins(
        midpoints=midpoints, 
        lb=bound_left, 
        ub=bound_right,
        channels_per_decade=int(meta['Channels/Decade'])
    )

    # rename the index
    data.index.rename("timestamp", inplace=True)

    if as_dict:
        return dict(
                meta=meta,
                units=units,
                weight=weight,
                data=data,
                bins=bins,
                bin_labels=bin_labels,
                bin_prefix='bin')
    else:
        return SMPS(data=data, meta=meta, bins=bins, 
                    bin_labels=bin_labels, units=units, weight=weight)


def load_sample(label="boston"):
    """Load a sample data file directly from GitHub.

    :param label: Which dataset to load.
    :type label: {'boston', 'chamber'}
    :return: The example data loaded from GitHub.
    :rtype: SMPS instance
    """
    assert(label in ["boston", "chamber"]), "Invalid option chosen \
    for the label"

    files = {
        'boston': {
            'uri': ("https://raw.githubusercontent.com/dhhagan/py-"
            + "smps/master/sample-data/boston_wintertime.txt"),
            'column': False
        },
        'chamber': {
            'uri': ("https://raw.githubusercontent.com/dhhagan/py-"
            + "smps/master/sample-data/mit_chamber_sample_column.txt"),
            'column': True
        }
    }

    m = smps_from_txt(fpath=files[label]['uri'], 
                      column=files[label]['column'])

    # convert to an SMPS instance
    return SMPS(
        data=m['data'], 
        bins=m['bins'], 
        meta=m['meta'], 
        bin_labels=m['bin_labels'], 
        weight=m['weight'], 
        units=m['units'])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
"""
import numpy as np
import pandas as pd
import math
import copy
import joblib
import json
from .utils import make_bins


__all__ = [
    "GenericParticleSizer", 
    "SMPS", 
    "AlphasenseOPCN2", 
    "AlphasenseOPCN3", 
    "ModulairPM", 
    "Modulair", 
    "POPS", 
    "ParticlesPlus", 
    "Grimm11D"
]


class GenericParticleSizer(object):
    """
        GenericParticleSizer is the base class for all
        available particle sizing instruments. It contains
        all of the basic functionality and methods used for
        making calculations.

        It can receive data in several formats including: 
        dN, dNdlogDp.
        
        :param data: Data values indexed by timestamp with 
            bins as columns.
        :type data: DataFrame
        :param bins: A list, ordered respective to the bins, 
            where for each bin a min, midpoint, 
            and max particle size is given.
        :type bins: list of lists
        :param dp_units: Units in which the diameter 
            is measured, defaults to 'um'.
        :type dp_units: {'um','nm'}
    """
    def __init__(self, data, bins, **kwargs):
        self.data = data.copy(deep=True)
        self.bins = bins
        self.meta = kwargs.pop('meta', dict())
        self.bin_labels = kwargs.pop('bin_labels', None)

        self.meta['weight'] = kwargs.pop("weight", "number")
        self.meta['units'] = kwargs.pop("units", "dw/dlogdp")

        self.bin_weights = kwargs.pop("bin_weights", 
                                      np.ones(len(self.bins)))

        # if bin_labels were not provided, try to generate them
        if self.bin_labels is None:
            self.bin_labels = [c for c in self.data.columns if 
                               kwargs.pop("bin_prefix", "bin") in c]

        # if no bin labels exist, raise an error
        assert(len(self.bin_labels) > 0), ("No bin labels have been "
        + "found or specified.")

        # make sure bins is has three columns
        assert(self.bins.shape[1] == 3), "`bins` must be an Nx3 array."

        # we always work with our units in um, so convert
        if kwargs.pop('dp_units', 'um') == 'nm':
            self.bins = self.bins * 1e-3

        # if the data is in a format other than 
        # [weight='number', units='dw/dlogdp']
        if kwargs.pop('fmt', 'dn') == 'dn':
            self.data[self.bin_labels] = (
                self
                .data[self.bin_labels]
                .div(self.dlogdp)
            )

        # multiply everything by the bin weights
        # if ones, this will have no effect
        self.data[self.bin_labels] = (
            self
            .data[self.bin_labels]
            .mul(self.bin_weights)
        )

    @property
    def s_multiplier(self):
        """
        A list of the surface areas corresponding
        to the midpoitns of the respective bins.
        """
        return 4 * np.pi * (self.bins[:, 1]/2)**2

    @property
    def v_multiplier(self):
        """
        A list of the volumes corresponding to the 
        midpoitns of the respective bins.
        """
        return (4./3)*np.pi*(self.bins[:, 1]/2)**3

    @property
    def midpoints(self):
        """
        A list of the midpoints of the respective bins.
        """
        return self.bins[:, 1]

    @property
    def dlogdp(self):
        """
        A list of the differences between the upper 
        and lower bound of each bin on a Log scale.
        """
        return np.log10(self.bins[:, -1]) - np.log10(self.bins[:, 0])

    @property
    def dn(self):
        """
        A DataFrame with the number of particles in 
        each bin at each time.
        """
        return self.dndlogdp.mul(self.dlogdp)

    @property
    def dndlogdp(self):
        """
        The histogram in units of dNd/logDp [#/cm3].
        """
        return self.data[self.bin_labels]

    @property
    def dddlogdp(self):
        """
        The histogram in units of dDpd/logDp [um/cm3].
        """
        return self.dndlogdp.mul(self.midpoints)

    @property
    def ds(self):
        """
        A DataFrame with the total surface area of 
        all the particles in each bin at each time.
        """
        return self.dn.mul(self.s_multiplier)

    @property
    def dsdlogdp(self):
        """
        The histogram in units of dSd/logDp [um2/cm3].
        """
        return self.dndlogdp.mul(self.s_multiplier)

    @property
    def dv(self):
        """
        A DataFrame with the total volume of all 
        the particels in each bin at each time.
        """
        return self.dvdlogdp.mul(self.dlogdp)

    @property
    def dvdlogdp(self):
        """
        The histogram in units of dSd/logDp [um3/cm3].
        """
        return self.dndlogdp.mul(self.v_multiplier)

    def copy(self):
        """
        Return a copy of the GenericParticleSizer.
        """
        return copy.deepcopy(self)

    def dump(self, filepath, ):
        """
        Save a copy to disk at `filepath`.

        :param filepath: The path at which to save the file.
        :type filepath: string
        :return: The list of file names in which the data is stored.
        :rtype: list of strings
        """
        return joblib.dump(self, filepath)

    @property
    def scan_stats(self):
        """
        A DataFrame with all the data except for the bin data.
        """
        return self.data[list(set(self.data.columns)
                              - set(self.bin_labels))]

    def _subselect_frame(self, df, dmin=0., dmax=1e3):
        """
        Sub-select the bins that are used to make calculations by
        zeroing out bins which are outside of the range. This
        is similar to a slice, except in the bin-dimension 
        rather than the time-dimension. Edge cases are 
        appropriately weighted.

        Units are in microns.
        
        :param df: A DataFrame with bins as columns
        :type df: DataFrame
        :param dmin: The minimum particle diameter in microns to 
            be included.
        :type dmin: float
        :param dmax: The maximum particle diameter in microns to 
            be included.
        :type dmax: float
        :return: A copy of `df` with the undesired columns
            zeroed out.
        :rtype: DataFrame
        """
        # generate an array of factors to multiply by
        factors = np.zeros(self.bins.shape[0])

        for i, b in enumerate(self.bins):
            if b[0] >= dmin and b[-1] <= dmax:
                factors[i] = 1
            if dmin >= b[0] and dmin < b[-1]:
                factors[i] = (b[-1] - dmin) / (b[-1] - b[0])
            if dmax >= b[0] and dmax < b[-1]:
                factors[i] = (dmax - b[0]) / (b[-1] - b[0])

        # copy the dataframe
        cpy = df.copy()

        # multiply the histogram by the factors we just calculated
        cpy[self.bin_labels] = cpy[self.bin_labels].mul(factors)

        return cpy

    def stats(self, weight='number', dmin=0., 
              dmax=1e3, rho=1.65, **kwargs):
        """
        Calculate and return the total number of particles, 
        total surface area, total volume, and total mass 
        between dmin and dmax. In addition, the arithmetic 
        mean (AM), geometric mean (GM), mode (Mode), and geometric 
        standard deviation (GSD, CMD) are calculated and are
        weighted by the input parameter `weight`.
        
        :param weight: The dimension to consider in computing 
            statistics.
        :type weight: {"number", "surface", "volume", "mass"}
        :param dmin: The minimum particle diameter in microns, 
            defaults to `0`.
        :type dmin: float
        :param dmax: The maximum particle diameter in microns, 
            defaults to `1e3`.
        :type dmax: float
        :param rho: The particle density in units of g/cc3, 
            defaults to `1.65`.
        :type rho: float
        :return: The statistics for each timestamp.
        :rtype: DataFrame
        """
        # remove the weight from the kwargs
        assert(weight in ["number", "surface", "volume", "mass"])

        # initialize an empty dataframe to hold the results
        res = pd.DataFrame(index=self.data.index)

        # subselect the dataframe to only include 
        # diameters of interest
        cpy = self.copy()
        cpy.data = cpy._subselect_frame(cpy.data, **kwargs)

        # drop all rows that are completely NaN
        cpy.data = cpy.data.dropna(how='all')

        # make a shortcut for the index to inject
        idx = cpy.data.index

        # calculate the total number of particles
        res.loc[idx, "number"] = cpy.dn.sum(axis=1)
        res.loc[idx, "surface_area"] = cpy.ds.sum(axis=1)
        res.loc[idx, "volume"] = cpy.dv.sum(axis=1)
        res.loc[idx, "mass"] = cpy.dv.mul(rho).sum(axis=1)

        if weight == "number":
            res.loc[idx, "AM"] = (
                1e3*cpy.dn.mul(self.midpoints).sum(axis=1) 
                / res.loc[idx, "number"]
            )
            res.loc[idx, "GM"] = 1e3*np.exp(
                cpy.dn.mul(np.log(self.midpoints), axis=1).sum(axis=1) 
                / res.loc[idx, "number"]
            )
            rename_cols_dict = dict(zip(
                cpy.dndlogdp.columns, 
                cpy.midpoints
            ))
            res.loc[idx, "Mode"] = 1e3 * (
                cpy
                .dndlogdp
                .rename(columns=rename_cols_dict)
                .idxmax(axis=1)
            )

            tmp = cpy.dn.assign(GM=res.loc[idx, 'GM'].values)
        elif weight == "surface":
            res.loc[idx, "AM"] = 1e3 * (
                cpy
                .ds
                .mul(self.midpoints)
                .sum(axis=1) 
                / res.loc[idx, "surface_area"]
            )
            res.loc[idx, "GM"] = 1e3 * np.exp(
                cpy
                .ds
                .mul(np.log(self.midpoints), axis=1)
                .sum(axis=1) 
                / res.loc[idx, "surface_area"]
            )
            rename_cols_dict = dict(zip(
                cpy.dsdlogdp.columns, 
                cpy.midpoints
            ))
            res.loc[idx, "Mode"] = 1e3 * (
                cpy
                .dsdlogdp
                .rename(columns=rename_cols_dict)
                .idxmax(axis=1)
            )

            tmp = cpy.ds.assign(GM=res.loc[idx, 'GM'].values)
        else:
            res.loc[idx, "AM"] = 1e3 * (
                cpy
                .dv
                .mul(self.midpoints)
                .sum(axis=1) 
                / res.loc[idx, "volume"]
            )
            res.loc[idx, "GM"] = 1e3 * np.exp(
                cpy
                .dv
                .mul(np.log(self.midpoints), axis=1)
                .sum(axis=1) 
                / res.loc[idx, "volume"]
            )
            rename_cols_dict = dict(zip(
                cpy.dvdlogdp.columns, 
                cpy.midpoints))
            res.loc[idx, "Mode"] = 1e3 * (
                cpy
                .dvdlogdp
                .rename(columns=rename_cols_dict)
                .idxmax(axis=1)
            )

            tmp = cpy.dv.assign(GM=res.loc[idx, 'GM'].values)
        
        if weight == "mass":
            res.loc[idx, "AM"] = res.loc[idx, "AM"] * rho
            res.loc[idx, "GM"] = res.loc[idx, "GM"] * rho
            res.loc[idx, "Mode"] = res.loc[idx, "Mode"] * rho
        
        
        # calculate the GSD
        res.loc[idx, "GSD"] = self._gsd(tmp)

        # delete the cpy to free up memory
        del cpy, tmp

        return res

    def integrate(self, weight='number', dmin=0., dmax=1., **kwargs):
        """
        For each timestamp give the total number 
        (or surface, volume, or mass) of all the particles 
        (with a diameter between `dmin` and `dmax`) at that time.
        
        :param weight: The variable to sum over, 
            defaults to "number".
        :type weight: {"number", "surface", "volume", "mass"}
        :param dmin: The minimum particle diameter in microns, 
            defaults to `0`.
        :type dmin: float
        :param dmax: The maximum particle diameter in microns, 
            defaults to `1e3`.
        :type dmax: float
        :param rho: The particle density in units of g/cc3, 
            defaults to `1.65`.
        :type rho: float
        :return: The desired values indexed by timestamp.
        :rtype: Pandas Series
        """
        rho = kwargs.pop("rho", 1.65)

        weight = weight.lower()

        assert(weight in ['number', 'surface', 'volume', 'mass']), \
        "Invalid `weight`"

        if weight == 'number':
            df = self.dn
        elif weight == 'surface':
            df = self.ds
        elif weight == 'volume':
            df = self.dv
        else:
            df = self.dv * rho

        # subsample the data
        df = self._subselect_frame(df, dmin=dmin, dmax=dmax)

        return df[self.bin_labels].sum(axis=1)

    def slice(self, start=None, end=None, inplace=False):
        """
        Slice the data between the start and end dates, inclusive.
        
        :param start: The first timestamp to be contained 
            in the result.
        :type start: string, Pandas Timestamp, or numpy DateTime64
        :param end: The last timestamp to be contained in the 
            result.
        :type end: string, Pandas Timestamp, or numpy DateTime64
        :param inplace: If False a new object is returned, 
            defaults to False.
        :type inplace: bool
        :return: The data restricted to the given time interval.
        :rtype: an object the same type as self
        """
        if inplace:
            self.data = self.data[start:end]
        else:
            cpy = self.copy()

            cpy.data = cpy.data[start:end]

            return cpy
        return

    def resample(self, rs, inplace=False):
        """
        Resample on the timeseries.

        :param rs: The resample period using normal 
            Pandas conventions.
        :type rs: string
        :param inplace: If True, update the current instance, 
            otherwise create a new one and return it.
        :type inpace: bool
        :return: A timeseries that has been resampled with 
            period `rs`.
        :rtype: an object from the same class as `self`

        Example::
            m.resample("15min", True)

        """
        obj_cols = (
            self
            .data
            .select_dtypes(include=['object'])
            .resample(rs)
            .first()
        )
        num_cols = (
            self
            .data
            .select_dtypes(exclude=['object'])
            .resample(rs)
            .mean()
        )

        # re-merge the two dataframes
        merged = pd.merge(num_cols, obj_cols, left_index=True, 
                          right_index=True, how='outer')

        if inplace:
            self.data = merged
        else:
            cpy = self.copy()
            cpy.data = merged

            return cpy

        return

    def _gsd(self, df):
        """
        Private function to calculate the geometric 
        standard deviation for a lognormal distribution.
        
        :param df: A DataFrame with a `GM` column holding 
            the geometric mean and columns of bins. 
            In each row the bin value will act as the 
            weight on the midpoint value for that bin, 
            found in `self.midpoints`. 
        :type df: DataFrame
        :return: A column holding the geometric mean for each row.
        :rtype: Pandas Series
        """

        # calculate the GM in units of microns
        gm = df['GM'] * 1e-3

        # drop GM. At this point, the row will have the 
        # histogram and nothing else
        df = df.drop("GM", axis=1)

        # calculate the geometric standard deviation
        gsd = np.exp(
            np.sqrt(
                df
                 .mul(
                     (df*0)
                     .add(np.log(self.midpoints), axis=1)
                     .sub(np.log(gm), axis=0)
                     **2
                 )
                 .sum(axis=1)
                / (df.sum(axis=1))
            )
        )

        return gsd

    
class SMPS(GenericParticleSizer):
    def __init__(self, **kwargs):
        super(SMPS, self).__init__(fmt='dndlogdp', dp_units='nm', 
                                   **kwargs)
    def __repr__(self):
        return "<SMPS>"


class AlphasenseOPCN2(GenericParticleSizer):
    """
    The Alphasense OPC-N2 is a consumer-grade optical 
    particle counter that uses a 658 nm red laser to count 
    and size individual particles. It can count particles 
    between 380-17,500 nm. Read more `here
    <https://www.alphasense.com/products/
    optical-particle-counter/>`__.
    """
    def __init__(self, **kwargs):
        bb = np.array([0.38, 0.54, 0.78, 1.05, 1.34, 1.59, 2.07, 3., 
                       4., 5., 6.5, 8., 10., 12., 14., 16., 17.5])

        bins = kwargs.pop("bins", make_bins(boundaries=bb))
        fmt = kwargs.pop("fmt", "dn")
        bin_labels = kwargs.pop(
            "bin_labels", 
            ["bin{}".format(i) for i in range(bins.shape[0])]
        )

        super(AlphasenseOPCN2, self).__init__(
            bins=bins, 
            fmt=fmt, 
            bin_labels=bin_labels, 
            **kwargs
        )


class AlphasenseOPCN3(GenericParticleSizer):
    """
    The Alphasense OPC-N3 is a consumer-grade optical
    particle counter that uses a 658 nm red laser to count 
    and size individual particles. It can count particles 
    between 380-40,000 nm. Read more `here <https://www.
    alphasense.com/products/optical-particle-counter/>`__.
    """
    def __init__(self, **kwargs):
        bb = np.array([0.35, 0.46, 0.66, 1., 1.3, 1.7, 2.3, 3.,
                            4., 5.2, 6.5, 8., 10., 12., 14., 16., 18.,
                            20., 22., 25., 28., 31., 34., 37., 40.])

        bins = kwargs.pop("bins", make_bins(boundaries=bb))
        fmt = kwargs.pop("fmt", "dn")
        bin_labels = kwargs.pop(
            "bin_labels", 
            ["bin{}".format(i) for i in range(bins.shape[0])]
        )

        super(AlphasenseOPCN3, self).__init__(
            bins=bins, 
            fmt=fmt, 
            bin_labels=bin_labels, 
            **kwargs
        )

class ModulairPM(GenericParticleSizer):
    """
    QuantAQ's MODULAIR-PM sensors use the Alphasense OPC-N3,
    a consumer-grade optical particle counter that uses a 658 nm 
    red laser to count and size individual particles. It can 
    count particles between 380-40,000 nm. Read more 
    `here <https://www.alphasense.com/products/optical-particle-counter/>`__.
    """
    def __init__(self, **kwargs):
        bb = np.array([0.35, 0.46, 0.66, 1., 1.3, 1.7, 2.3, 3.,
                            4., 5.2, 6.5, 8., 10., 12., 14., 16., 18.,
                            20., 22., 25., 28., 31., 34., 37., 40.])

        bins = kwargs.pop("bins", make_bins(boundaries=bb))
        fmt = kwargs.pop("fmt", "dn")
        bin_labels = kwargs.pop(
            "bin_labels", 
            ["bin{}".format(i) for i in range(bins.shape[0])]
        )

        super(ModulairPM, self).__init__(
            bins=bins, 
            fmt=fmt, 
            bin_labels=bin_labels, 
            **kwargs
        )

class Modulair(GenericParticleSizer):
    """
    QuantAQ's MODULAIR sensors use the Alphasense OPC-N3,
    a consumer-grade optical particle counter that uses a 658 nm 
    red laser to count and size individual particles. It can 
    count particles between 380-40,000 nm. However, because the last
    few bins (bin18-bin24) collect negligible data, they are removed to conserve 
    memory, which means the MODULAIR data differs slightly from the MODULAIR-PM data.
    Read more about the Alphasense OPC-N3
    `here <https://www.alphasense.com/products/optical-particle-counter/>`__.
    """
    def __init__(self, **kwargs):
        bb = np.array([0.35, 0.46, 0.66, 1., 1.3, 1.7, 2.3, 3.,
                            4., 5.2, 6.5, 8., 10., 12., 14., 16., 18.,
                            20., 22.,])

        bins = kwargs.pop("bins", make_bins(boundaries=bb))
        fmt = kwargs.pop("fmt", "dn")
        bin_labels = kwargs.pop(
            "bin_labels", 
            ["bin{}".format(i) for i in range(bins.shape[0])]
        )

        super(Modulair, self).__init__(
            bins=bins, 
            fmt=fmt, 
            bin_labels=bin_labels, 
            **kwargs
        )


class POPS(GenericParticleSizer):
    """
    The Portable Optical Particle Spectrometer is based on 
    the design of R. Gao and is currently manufactured by 
    Handix Scientific. It is a research-grade OPC that can see 
    particles as small as 132 nm. Read more `here 
    <http://www.handixscientific.com/pops>`__.
    """
    def __init__(self, **kwargs):
        # default left and right bounds according to Handix
        bb = 1e-3*np.array([132, 144, 158, 174, 191, 209, 229, 270, 324, 
                            473, 594, 1009, 1294, 1637, 2148, 2864, 3648])

        # make the 3xn array of bins using the utility 
        # function make_bins()
        bins = kwargs.pop('bins', make_bins(boundaries=bb))

        #
        fmt = kwargs.pop("fmt", "dn")

        super(POPS, self).__init__(
            bins=bins, 
            fmt=fmt, 
            bin_labels=bin_labels, 
            **kwargs
        )


class ParticlesPlus(GenericParticleSizer):
    """
    The Particles Plus is a consumer-grade optical particle counter
    that uses a red laser to count and size individual particles. It
    can count particles between 300-10,000 nm. Read more `here 
    <https://particlesplus.com/ambient-air-monitoring/>`__.
    """
    def __init__(self, **kwargs):
        bb = np.array([0.3, 0.5, 0.7, 1., 2.5, 10.0, 10.1])

        bins = kwargs.pop("bins", make_bins(boundaries=bb))
        fmt = kwargs.pop("fmt", "dn")
        bin_labels = kwargs.pop(
            "bin_labels", 
            ["bin{}".format(i) for i in range(bins.shape[0])]
        )

        super(ParticlesPlus, self).__init__(
            bins=bins, 
            fmt=fmt, 
            bin_labels=bin_labels, 
            **kwargs
        )


class Grimm11D(GenericParticleSizer):
    """
    The Grimm11D is an optical particle counter that uses a blue 
    laser to count and size individual particles. It can count 
    particles between 250-35,000 nm. Read more `here 
    <https://www.grimm-aerosol.com/products-en/dust-monitors/the-
    dust-decoder/11-d/>`__.
    """
    def __init__(self, **kwargs):
        bb = np.array([
            0.25, 0.28, 0.3, 0.35, 0.4, 0.45, 0.5, 0.58, 0.65, 0.7, 
            0.8, 1.0, 1.3, 1.6, 2.0, 2.5, 3.0, 3.5, 4., 5., 6.5, 7.5, 
            8.5, 10., 12.5, 15., 17.5, 20., 25., 30., 32., 35.
        ])

        bins = kwargs.pop("bins", make_bins(boundaries=bb))
        fmt = kwargs.pop("fmt", "dn")
        bin_labels = kwargs.pop(
            "bin_labels", 
            ["bin{}".format(i) for i in range(bins.shape[0])]
        )

        super(Grimm11D, self).__init__(
            bins=bins, 
            fmt=fmt, 
            bin_labels=bin_labels, 
            **kwargs
        )



"""Functions to support fitting multi-modal distributions.
"""
import numpy as np
from scipy.optimize import curve_fit
from statsmodels.iolib.table import SimpleTable
from .utils import Table

def dndlogdp(dp, n, gm, gsd):
    """
    Discrete PDF for a lognormal particle size distribution.
    
    :param dp: Particle diameters at which to evaluate the 
        PDF in building the discrete model.
    :type dp: numpy array
    :param n: Area under the curve of the continuous PDF 
        (the discrete model will be slightly different).
    :type n: float
    :param gm: The geometric mean.
    :type gm: float
    :param gsd: The geometric standard deviation.
    :type gsd: float
    :return: A list of concentrations respective to the 
        list of particle sizes, `dp`.
    :rtype: numpy array
    """
    return  ((n/(np.sqrt(2*np.pi) * np.log10(gsd))) * 
             np.exp(-((np.log10(dp) - np.log10(gm))**2) / 
                    (2*np.log10(gsd)**2)))

def dsdlogdp(dp, n, gm, gsd):
    """
    Discrete PDF for a surface-weighted lognormal 
    particle size distribution.
    
    :param dp: Particle diameters at which to evaluate 
        the PDF in building the discrete model.
    :type dp: numpy array
    :param n: The area under the curve of the continuous PDF 
        (the discrete model will be slightly different).
    :type n: float
    :param gm: The geometric mean.
    :type gm: float
    :param gsd: The geometric standard deviation.
    :type gsd: float
    :return: A list of concentrations respective to the 
        list of particle sizes, `dp`.
    :rtype: numpy array
    """
    return np.pi * dp**2 * dndlogdp(dp, n, gm, gsd)

def dvdlogdp(dp, n, gm, gsd):
    """
    Discrete PDF for a volume-weighted lognormal particle 
    size distribution.
    
    :param dp: Particle diameters at which to evaluate 
        the PDF in building the discrete model.
    :type dp: numpy array
    :param n: The area under the curve of the continuous PDF 
        (the discrete model will be slightly different).
    :type n: float
    :param gm: The geometric mean.
    :type gm: float
    :param gsd: The geometric standard deviation.
    :type gsd: float
    :return: A list of concentrations respective to 
        the list of particle sizes, `dp`.
    :rtype: numpy array
    """
    return (np.pi/6.) * (dp**3) * dndlogdp(dp, n, gm, gsd)

def number_weighted_single_mode(dp, n, gm, gsd):
    """
    Discrete PDF for a lognormal distribution of particle 
    size for a 1-mode distribution.
    
    :param dp: Particle diameters at which to evaluate 
        the PDF in building the discrete model.
    :type dp: numpy array
    :param n: The  area under the curve of the continuous PDF 
        (the discrete model will be slightly different).
    :type n: float
    :param gm: The geometric mean.
    :type gm: float
    :param gsd: The geometric standard deviation.
    :type gsd: float
    :return: A list of concentrations respective to the 
        list of particle sizes, `dp`.
    :rtype: numpy array
    """
    return dndlogdp(dp, n, gm, gsd)

def number_weighted_two_modes(dp, n1, gm1, gsd1, n2, gm2, gsd2):
    """
    Discrete PDF for a lognormal distribution of particle 
    size for a 2-mode distribution.
    
    :param dp: Particle diameters at which to evaluate 
        the PDF in building the discrete model.
    :type dp: numpy array
    :param n1: The area under the curve of the continuous 
        PDF for the first component (the discrete 
        model will be slightly different).
    :type n1: float
    :param gm1: The geometric mean for the first component.
    :type gm1: float
    :param gsd1: The geometric standard deviation for 
        the first component.
    :type gsd1: float
    :param n2: The area under the curve of the continuous 
        PDF for the second component (the discrete 
        model will be slightly different).
    :type n2: float
    :param gm2: The geometric mean for the second component.
    :type gm2: float
    :param gsd2: The geometric standard deviation for 
        the second component.
    :type gsd2: float
    :return: A list of concentrations respective to 
        the list of particle sizes, `dp`.
    :rtype: numpy array
    """
    N = [n1, n2]
    GM = [gm1, gm2]
    GSD = [gsd1, gsd2]

    s = 0
    for i in range(len(N)):
        s += dndlogdp(dp, N[i], GM[i], GSD[i])

    return s

def number_weighted_three_modes(dp, n1, gm1, gsd1, n2, gm2, 
                                gsd2, n3, gm3, gsd3):
    """
    Discrete PDF for a lognormal distribution of particle 
        size for a 3-mode distribution.
    
    :param dp: The particle diameters at which to evaluate 
        the PDF in building the discrete model.
    :type dp: numpy array
    :param n1: The area under the curve of the continuous 
        PDF for the first component (the discrete 
        model will be slightly different).
    :type n1: float
    :param gm1: The geometric mean for the first component.
    :type gm1: float
    :param gsd1: The geometric standard deviation for .
        the first component
    :type gsd1: float
    :param n2: The area under the curve of the continuous 
        PDF for the second component (the discrete 
        model will be slightly different).
    :type n2: float
    :param gm2: The geometric mean for the second component.
    :type gm2: float
    :param gsd2: The geometric standard deviation for 
        the second component.
    :type gsd2: float
    :param n3: The area under the curve of the continuous 
        PDF for the third component (the discrete 
        model will be slightly different).
    :type n3: float
    :param gm3: The geometric mean for the third component.
    :type gm3: float
    :param gsd3: The geometric standard deviation for 
        the third component.
    :type gsd3: float
    :return: A list of concentrations respective to 
        the list of particle sizes, `dp`.
    :rtype: numpy array
    """
    N = [n1, n2, n3]
    GM = [gm1, gm2, gm3]
    GSD = [gsd1, gsd2, gsd3]

    s = 0
    for i in range(len(N)):
        s += dndlogdp(dp, N[i], GM[i], GSD[i])

    return s

def surface_weighted_single_mode(dp, n, gm, gsd):
    """
    Discrete PDF for a lognormal distribution of particle 
    size for a 1-mode distribution weighted by surface.
    
    :param dp: The particle diameters at which to evaluate 
        the PDF in building the discrete model.
    :type dp: numpy array
    :param n: The area under the curve of the continuous 
        PDF (the discrete model will be slightly different).
    :type n: float
    :param gm: The geometric mean.
    :type gm: float
    :param gsd: The geometric standard deviation.
    :type gsd: float
    :return: A list of concentrations respective to 
        the list of particle sizes, `dp`.
    :rtype: numpy array
    """
    return dsdlogdp(dp, n, gm, gsd)

def surface_weighted_two_modes(dp, n1, gm1, gsd1, n2, gm2, gsd2):
    """
    Discrete PDF for a lognormal distribution of particle 
    size for a 2-mode distribution weighted by surface.
    
    :param dp: The particle diameters at which to evaluate 
        the PDF in building the discrete model.
    :type dp: numpy array
    :param n1: The area under the curve of the continuous PDF 
        for the first component (the discrete model will 
        be slightly different).
    :type n1: float
    :param gm1: The geometric mean for the first component.
    :type gm1: float
    :param gsd1: The geometric standard deviation for 
        the first component.
    :type gsd1: float
    :param n2: The area under the curve of the continuous 
        PDF for the second component (the discrete model 
        will be slightly different).
    :type n2: float
    :param gm2: The geometric mean for the second component.
    :type gm2: float
    :param gsd2: The geometric standard deviation for the 
        second component.
    :type gsd2: float
    :return: A list of concentrations respective to the 
        list of particle sizes, `dp`.
    :rtype: numpy array
    """
    N = [n1, n2]
    GM = [gm1, gm2]
    GSD = [gsd1, gsd2]

    s = 0
    for i in range(len(N)):
        s += dsdlogdp(dp, N[i], GM[i], GSD[i])

    return s

def surface_weighted_three_modes(dp, n1, gm1, gsd1, n2, gm2, 
                                 gsd2, n3, gm3, gsd3):
    """
    Discrete PDF for a lognormal distribution of particle size 
    for a 3-mode distribution weighted by surface.
    
    :param dp: The particle diameters at which to evaluate 
        the PDF in building the discrete model.
    :type dp: numpy array
    :param n1: The area under the curve of the continuous PDF 
        for the first component (the discrete model will 
        be slightly different).
    :type n1: float
    :param gm1: The geometric mean for the first component.
    :type gm1: float
    :param gsd1: The geometric standard deviation for 
        the first component.
    :type gsd1: float
    :param n2: The area under the curve of the continuous PDF 
        for the second component (the discrete model 
        will be slightly different).
    :type n2: float
    :param gm2: The geometric mean for the second component.
    :type gm2: float
    :param gsd2: The geometric standard deviation for 
        the second component.
    :type gsd2: float
    :param n3: The area under the curve of the continuous PDF 
        for the third component (the discrete model will be 
        slightly different).
    :type n3: float
    :param gm3: The geometric mean for the third component.
    :type gm3: float
    :param gsd3: The geometric standard deviation for the 
        third component.
    :type gsd3: float
    :return: A list of concentrations respective to the 
        list of particle sizes, `dp`.
    :rtype: numpy array
    """
    N = [n1, n2, n3]
    GM = [gm1, gm2, gm3]
    GSD = [gsd1, gsd2, gsd3]

    s = 0
    for i in range(len(N)):
        s += dsdlogdp(dp, N[i], GM[i], GSD[i])

    return s

def volume_weighted_single_mode(dp, n, gm, gsd):
    """
    Discrete PDF for a lognormal distribution of particle 
    size for a 1-mode distribution weighted by volume.
    
    :param dp: The particle diameters at which to evaluate the 
        PDF in building the discrete model.
    :type dp: numpy array
    :param n: The area under the curve of the continuous PDF 
        (the discrete model will be slightly different).
    :type n: float
    :param gm: The geometric mean.
    :type gm: float
    :param gsd: The geometric standard deviation.
    :type gsd: float
    :return: A list of concentrations respective to the 
        list of particle sizes, `dp`.
    :rtype: numpy array
    """
    return dvdlogdp(dp, n, gm, gsd)

def volume_weighted_two_modes(dp, n1, gm1, gsd1, n2, gm2, gsd2):
    """
    Discrete PDF for a lognormal distribution of particle 
    size for a 2-mode distribution weighted by volume.
    
    :param dp: The particle diameters at which to evaluate the 
        PDF in building the discrete model.
    :type dp: numpy array
    :param n1: The area under the curve of the continuous PDF 
        for the first component (the discrete model will 
        be slightly different).
    :type n1: float
    :param gm1: The geometric mean for the first component.
    :type gm1: float
    :param gsd1: The geometric standard deviation for the 
        first component.
    :type gsd1: float
    :param n2: The area under the curve of the continuous PDF 
        for the second component (the discrete model will be 
        slightly different).
    :type n2: float
    :param gm2: The geometric mean for the second component.
    :type gm2: float
    :param gsd2: The geometric standard deviation for the 
        second component.
    :type gsd2: float
    :return: A list of concentrations respective to the 
        list of particle sizes, `dp`.
    :rtype: numpy array
    """
    N = [n1, n2]
    GM = [gm1, gm2]
    GSD = [gsd1, gsd2]

    s = 0
    for i in range(len(N)):
        s += dvdlogdp(dp, N[i], GM[i], GSD[i])

    return s

def volume_weighted_three_modes(dp, n1, gm1, gsd1, n2, gm2, gsd2,
                                n3, gm3, gsd3):
    """
    Discrete PDF for a log-normal distribution of particle 
    size for a 3-mode distribution weighted by volume.
    
    :param dp: The particle diameters at which to evaluate the 
        PDF in building the discrete model.
    :type dp: numpy array
    :param n1: The area under the curve of the continuous PDF 
        for the first component (the discrete model will be 
        slightly different).
    :type n1: float
    :param gm1: The geometric mean for the first component.
    :type gm1: float
    :param gsd1: The geometric standard deviation for the 
        first component.
    :type gsd1: flaot
    :param n2: The area under the curve of the continuous PDF 
        for the second component (the discrete model will 
        be slightly different).
    :type n2: float
    :param gm2: The geometric mean for the second component.
    :type gm2: float
    :param gsd2: The geometric standard deviation for the 
        second component.
    :type gsd2: float
    :param n3: The area under the curve of the continuous PDF
        for the third component (the discrete model will be
        slightly different).
    :type n3: float
    :param gm3: The geometric mean for the third component.
    :type gm3: float
    :param gsd3: The geometric standard deviation for the
        third component.
    :type gsd3: float
    :return: A list of concentrations respective to the list
        of particle sizes, `dp`.
    :rtype: numpy array
    """
    N = [n1, n2, n3]
    GM = [gm1, gm2, gm3]
    GSD = [gsd1, gsd2, gsd3]

    s = 0
    for i in range(len(N)):
        s += dvdlogdp(dp, N[i], GM[i], GSD[i])

    return s

models = {
    'number': [
        number_weighted_single_mode,
        number_weighted_two_modes,
        number_weighted_three_modes
        ],
    'surface': [
        surface_weighted_single_mode,
        surface_weighted_two_modes,
        surface_weighted_three_modes
        ],
    'volume': [
        volume_weighted_single_mode,
        volume_weighted_two_modes,
        volume_weighted_three_modes
        ]
    }

class LogNormal(object):
    """
    `LogNormal` uses non-linear least squares to fit a
    log-normal model, with up to three modes.
    """
    def __init__(self):
        pass

    def fit(self, X, Y, modes=1, xmin=None,
            xmax=None, weight='number', fit_kwargs=None, **kwargs):
        """
        `fit` uses non-linear least squares to fit a
        log-normal model, with up to three modes.
        
        :param X: The independant variable.
        :type X: array-like
        :param Y: The dependent variable.
        :type Y: array-like
        :param modes: Number of modes for the model to 
            have, defaults to 1.
        :type modes: {1,2,3}
        :param xmin: Lower bound by which to crop data.
        :type xmin: float, optional
        :param xmax: Upperbound by which to crop data.
        :type xmax: float, optional
        :param weight: Dimension of interest.
        :type weight: {"number", "surface", "volume"}
        :param fit_kwargs: Options to be passed 
            to ``scipy.optimize.curve_fit``. Read 
            more `here <https://docs.scipy.org/doc/
            scipy/reference/generated/scipy.optimize.
            curve_fit.html>`__.
        :type fit_kwargs: dict, optional
        :param p0: Initial guesses passed to ``scipy.
            optimize.curve_fit`` for the parameters. 
            Must have length 3*`modes`.
        :type p0: list, optional
        :param bounds: Bounds to be passed to ``scipy.
            optimize.curve_fit``. Defaults to `(0, [1e9, 
            1e5, 5]*self.modes)`.
        :type bounds: 2-tuple
        :return: A fitted log-normal model.
        :rtype: `LogNormalFitResults` object
        """
        
        self.modes = modes

        if fit_kwargs is None:
            fit_kwargs = dict()

        p0 = kwargs.pop('p0', [1e5, 1.5, 2, 1e5, 5, 2.5, 1e3, 50, 2.5])

        bounds = kwargs.pop('bounds', (0, [1e9, 1e5, 5]*self.modes))

        self.model = models[weight][self.modes-1]

        # Set the initial guesses
        p0 = p0[0:self.modes*3]

        # Subset the data if xmin or xmax is set
        if xmin is not None:
            X = X[np.where(X >= xmin)]
            Y = Y[np.where(X >= xmin)]

        if xmax is not None:
            X = X[np.where(X <= xmax)]
            Y = Y[np.where(X <= xmax)]

        self.fit_params, cov = curve_fit(self.model, X, Y, p0,
                                bounds=bounds, **fit_kwargs)

        perr = np.sqrt(np.diag(cov))

        fittedvalues = self.model(X, *self.fit_params)

        return LogNormalFitResults(params=self.fit_params,
                                   error_matrix=perr,
                                   fittedvalues=fittedvalues,
                                   modes=self.modes)


class LogNormalFitResults(object):
    """
    `LogNormalFitResults` holds the results produced by `LogNormal`.
    
    :param params: The parameters of the fitted model.
    :type params: array
    :param error_matrix: The one standard deviation errors associaed 
        with the paramters, given in the same order as `params`. 
    :type error_matrix: array
    :param fittedvalues: The results from evaluating the model at 
        each of the independent variables in the training data.
    :type fittedvalues: array
    :param modes: The number of modes the model has.
    :type modes: {1,2,3}
    """
    def __init__(self, params, error_matrix, fittedvalues,
                 modes, **kwargs):
        self.modes = modes

        self.params = params.reshape(self.modes, 3)
        self.errors = error_matrix.reshape(self.modes, 3)
        self.fittedvalues = fittedvalues

    def summary(self):
        """
        A `Table` object with the a summary of the model.
        """
        # Convert GM from microns to nm
        params = self.params.copy()
        errors = self.errors.copy()

        params[:, 1] *= 1000.
        errors[:, 1] *= 1000.

        _tbl = Table()

        _tbl.add_title(self.__class__.__name__)
        _tbl.add_border("=")
        _tbl.add_header()
        _tbl.add_border("-")

        # Add a row for each mode
        for i in range(self.modes):
            _tbl.add_row(label="Mode {}".format(i),
                         fields=params[i], errors=errors[i])

        _tbl.add_border("-")

        return _tbl

    def predict(self, X, weight='number'):
        """
        Uses the model to predict the outcome for a given input.
        
        :param X: Particle diameters at which to evaluate 
            the model PDF.
        :type X: numpy array
        :param weight: Dimension of concentration to model.
        :type weight: {"number", "surface", "volume"}
        :return: A list of concentrations respective to the 
            list of particle sizes, `X`.
        :rtype: numpy array
        """
        model = models[weight][self.modes-1]

        return model(X, *self.params.flatten())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
"""
import numpy as np
import requests
import math
from scipy.stats.mstats import gmean

def make_bins(**kwargs):
    """
    'make_bins` computes the upper and lower bounds and
    midpoints of all the bins given either
    * `boundaries`
    * `boundaries_left` and `boundaries_right`
    * or `lb`, `ub`, and `midpoints` (`channels_per_decade` optional).
    
    :param boundaries: A list containing, for each bin, a list of 
        the lower and upper bound for the bin.
    :type boundaries: list of lists, or numpy array
    :param boundaries_left: A list of lower boundaries for each bin.
    :type boundaries_left: list
    :param boundaries_right: A list of upper boundaries for each 
        bin.
    :type boundaries_right: list
    :param lb: The lower bound of the lowermost bin.
    :type lb: float
    :param ub: The upper bound of the uppermost bin.
    :type ub: float
    :param midpoints: A list of midpoints for each bin.
    :type midpoints: list
    :param channels_per_decade: A measure of bin width on the log 
        scale.
    :type channels_per_decade: float
    :param mean_calc: Determins wether the geometric or arethmetic 
        mean should be used to compute the midpoints.
    :type mean_calc: {"gm", "am"}
    :return: For each bin a row is given with the lower boundary, 
        the midpoint, and the upper bounary for that bin.
    :rtype: numpy array
    """
    boundaries = kwargs.pop("boundaries", None)
    boundaries_left = kwargs.pop("boundaries_left", None)
    boundaries_right = kwargs.pop("boundaries_right", None)
    lb = kwargs.pop("lb", None)
    ub = kwargs.pop("ub", None)
    midpoints = kwargs.pop("midpoints", None)
    cpd = kwargs.pop("channels_per_decade", 64)
    mean_calc = kwargs.pop("mean_calc", "am")

    # initialize bins
    bins = None

    if midpoints is not None:
        if lb is None or ub is None:
            raise Exception("A lower and upper bound must be set")

        bins = np.empty((midpoints.shape[0], 3))
        bins.fill(np.NaN)

        # fill the midpoints
        bins[:, 1] = midpoints

        # fill the bounds
        bins[0, 0] = lb
        bins[-1, -1] = ub

        # iterate and calculate the bounds
        for i in range(bins.shape[0] - 1):
            bins[i, 2] = round(
                math.pow(10, np.log10(bins[i, 0]) + 1./cpd), 
                4
            )
            bins[i+1, 0] = bins[i, 2]

        return bins

    if boundaries is not None:
        bins = np.empty((boundaries.shape[0]-1, 3))
        bins.fill(np.NaN)

        bins[:, 0] = boundaries[0:-1]
        bins[:, 2] = boundaries[1:]

    elif boundaries_left is not None:
        if boundaries_right is None:
            raise Exception("Missing attribute: `boundaries_right`")

        assert(boundaries_left.shape[0] == boundaries_right.shape[0]), \
            "Boundaries must be the same dimensions."

        bins = np.empty((boundaries_left.shape[0], 3))
        bins.fill(np.NaN)

        bins[:, 0] = boundaries_left
        bins[:, 2] = boundaries_right

    else:
        raise Exception("Not enough information to compute.")

    # calculate the midpoints
    assert(mean_calc in ("gm", "am")), "Invalid mean calculation method."

    if mean_calc == 'am':
        bins[:, 1] = (bins[:, 0] + bins[:, 2]) / 2
    else:
        bins[:, 1] = [gmean([x, y]) for x, y in zip(
            bins[:, 0], 
            bins[:, 2]
        )]

    return bins

def _get_bin_count(fpath, delimiter=',', encoding='ISO-8859-1'):
    """
    Gets the number of bins in the file.
    
    :param fpath: A path or url for the file (if a url it must 
        include `http`, and if a file path it must not contain 
        `http`).
    :type fpath: string
    :param delimiter: The delimiter between items in the file, 
        defaults to ','.
    :type delimiter: string
    :param encoding: The encoding for the file, defaults to 
        'ISO-8859-1'.
    :type encoding: string
    :return: The number of bins in the file.
    :rtype: int
    """
    bins = 0
    if 'http' in fpath:
        req = requests.get(fpath)

        for line in req.iter_lines():
            try:
                if float(line.decode(encoding).split(delimiter)[0]):
                    bins += 1
            except: pass
    else:
        with open(fpath, 'r', encoding=encoding) as f:
            for line in f:
                try:
                    if float(line.split(delimiter)[0]):
                        bins += 1
                except: pass

    return bins

def _get_linecount(fpath, keyword, delimiter=',', encoding='ISO-8859-1'):
    """
    Return the line number in a file where the first item is 
    `keyword`. If there is no such line, it will return the total 
    number of lines in the file.
    
    :param fpath: A path or url for the file (if a url it must 
        include `http`, and if a file path it must not 
        contain `http`).
    :type fpath: string
    :param keyword: The string to look for.
    :type keyword: string
    :param delimiter: The delimiter between items in the file, 
        defaults to ','.
    :type delimiter: string
    :param encoding: The encoding for the file, defaults to 
        'ISO-8859-1'.
    :type encoding: string
    :return: The line number in the file where the first item is 
        `keyword`.
    :rtype: int
    """
    linecount = 0

    if 'http' in fpath:
        req = requests.get(fpath)

        for line in req.iter_lines():
            startswith = line.decode(encoding).split(delimiter)[0]

            if startswith == keyword:
                break

            linecount += 1
    else:
        with open(fpath, 'r', encoding=encoding) as f:
            for line in f:
                startswith = line.split(delimiter)[0]

                if startswith == keyword:
                    break

                linecount += 1

    return linecount

def roundup(x):
    """
    round up to a multiple of 100
    
    :param x: The number to be rounded up.
    :type x: float
    :return: `x` rounded up to a multiple of 100.
    :rtype: float
    """
    return x if x % 100 == 0 else x + 100 - x % 100




class Table(object):
    """
    A table which can be constructed and then printed into 
    the terminal.
    
    :param max_width: The maximum width, in characters, of the 
        table to be printed in the terminal, defaults to 80.
    :type max_width: int
    """
    def __init__(self, max_width=80, **kwargs):
        self.text = ""
        self.max_width_chars = max_width

    def add_title(self, title):
        """
        Add a title to the top of the table.
        
        :param title: The title to be added.
        :type title: string
        """
        self.text = title.center(self.max_width_chars) + self.text

        return

    def add_border(self, char='='):
        """
        Add a border to the bottom of the table to visually 
        delimit between data.
        
        :param char: The character to be used for the border.
        :type char: string
        """
        assert(len(char)==1), "`char` must be a single character"
        self.text += '\n' + char*self.max_width_chars

        return

    def _center_text(self, text, width=None):
        """
        Pad text on each side with spaces so that it takes up 
        `width` characters.
        
        :param text: The text to center.
        :type text: string
        :param width: The width, in characters, in which to center 
            the text, defaults to a quarter of `max_width`.
        :type width: int
        :return: The padded text.
        :rtype: string        
        """
        if width is None:
            width = int(self.max_width_chars / 4)

        return text.center(width)

    def add_header(self):
        """
        Appends a header to the bottom of the table with four 
        sections: blank, "N (cm-3)", "GM (nm)", and "GSD".
        """
        self.text += (
            "\n" 
            + self._center_text("") 
            + self._center_text("N (cm-3)")
        )
        self.text += (
            self._center_text("GM (nm)") 
            + self._center_text("GSD")
        )

        return

    def add_row(self, label, fields, errors):
        self.text += "\n" + self._center_text(label)

        field1 = "{:.2e} ({:.1e})".format(fields[0], errors[0])
        field2 = "{:.2f} ({:.1e})".format(fields[1], errors[1])
        field3 = "{:.2f} ({:.1e})".format(fields[2], errors[2])

        self.text += (
            self._center_text(field1) 
            + self._center_text(field2) 
            + self._center_text(field3)
        )

        return

    def __repr__(self):
        return self.text



""""""

import matplotlib as mpl

__all__ = ["set"]

def set(tick_scale=1, rc=dict()):
    """
    Control plot style and scaling using seaborn and the 
    matplotlib rcParams interface.
    
    :param tick_scale: A scaler number controling the spacing 
        on tick marks, defaults to 1.
    :type tick_scale: float
    :param rc: Additional settings to pass to rcParams.
    :type rc: dict
    """
    rc_log_defaults = {
        'xtick.major.size': 10. * tick_scale,
        'xtick.minor.size': 6. * tick_scale,
        'ytick.major.size': 10. * tick_scale,
        'ytick.minor.size': 6. * tick_scale,
        'xtick.color': '0.0',
        'ytick.color': '0.0',
        'axes.linewidth': 1.75,
        'mathtext.default': 'regular'
    }

    mpl.rcParams.update(dict(rc_log_defaults, **rc))



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': '.'}

packages = \
['smps']

package_data = \
{'': ['*']}

install_requires = \
['joblib>=1.0,<2.0',
 'numpy>=1.23.2,<2.0.0',
 'pandas>=1.4,<2.0',
 'requests>=2.24,<3.0',
 'scipy>=1.9,<2.0',
 'seaborn>=0.10,<0.11',
 'setuptools>=65.3,<66.0',
 'statsmodels>=0.13,<0.14']

setup_kwargs = {
    'name': 'py-smps',
    'version': '2.1.0a0',
    'description': 'A simple python library to import and visualize data from particle sizing instruments.',
    'long_description': '[![PyPI version](https://badge.fury.io/py/py-smps.svg)](https://badge.fury.io/py/py-smps)\n[![Build Status](https://travis-ci.org/dhhagan/py-smps.svg?branch=master)](https://travis-ci.org/dhhagan/py-smps)\n[![Coverage Status](https://coveralls.io/repos/github/dhhagan/py-smps/badge.svg?branch=master)](https://coveralls.io/github/dhhagan/py-smps?branch=master)\n\n\n# py-smps\nPython library for the analysis and visualization of data from a Scanning Mobility Particle Sizer (SMPS) and other particle sizing instruments (SEMS, OPC\'s).\n\n## Dependencies\n\nThe full list of dependencies can be found in the `pyproject.toml` file and are summarized below:\n\n```py\npython = ">=3.8, <3.12"\nstatsmodels = "^0.13"\nseaborn = "^0.10"\njoblib = "^1.0"\nrequests = "^2.24"\nscipy = "^1.9"\nnumpy = "^1.23.2"\npandas = "^1.4"\n```\n\nAs of `v1.2.0a0`, the library should be compatible with Apple silicone (tested on both M1 and M2).\n\n## Python Versions\n\nPython3.8 through Python3.11 are currently supported.\n\n## Installation\n\nTo install from PyPi:\n\n    $ pip install py-smps [--upgrade]\n\nIf you\'d like the latest pre-release:\n\n    $ pip install py-smps --pre [--upgrade]\n\nTo install the edge release directly from GitHub:\n\n    pip install git+https://github.com/quant-aq/py-smps.git\n\n## Unittests\n\nUnittests can be run by issuing the following command from within the main repo:\n\n```sh\n$ poetry run pytest -s tests/ --ignore=tests/datafiles\n```\n\n\n## Documentation\n\nDocumentation is available [here](https://quant-aq.github.io/py-smps/). Docs are built using Sphinx and can be built locally by doing the following:\n\n```sh\n$ cd docs/\n$ make clean\n$ make guides\n$ make html\n$ cd ..\n```\n\nThen, you can navigate to your local directory at `docs/_build/html/` and open up the `index.html` file in your preferred browser window.\n\n\n## Contributing to Development\n\nWe welcome all contributions from the community in the form of issues reporting, feature requests, bug fixes, etc.\n\nIf there is a feature you would like to see or a bug you would like to report, please open an issue. We will try to get to things as promptly as possible. Otherwise, feel free to send PR\'s!\n\n\n## Colorbar Information\n\n  * [matplotlib colorbars](http://matplotlib.org/examples/color/colormaps_reference.html)\n  * [seaborn color palette](http://seaborn.pydata.org/tutorial/color_palettes.html)\n\n',
    'author': 'David H Hagan',
    'author_email': 'david.hagan@quant-aq.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/quant-aq/py-smps',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.12',
}


setup(**setup_kwargs)
